% This file is part of the Attempto Parsing Engine (APE).
% Copyright 2008-2010, Kaarel Kaljurand <kaljurand@gmail.com>.
%
% The Attempto Parsing Engine (APE) is free software: you can redistribute it and/or modify it
% under the terms of the GNU Lesser General Public License as published by the Free Software
% Foundation, either version 3 of the License, or (at your option) any later version.
%
% The Attempto Parsing Engine (APE) is distributed in the hope that it will be useful, but WITHOUT
% ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
% PURPOSE. See the GNU Lesser General Public License for more details.
%
% You should have received a copy of the GNU Lesser General Public License along with the Attempto
% Parsing Engine (APE). If not, see http://www.gnu.org/licenses/.


:- module(implication_to_swrl, [
		implication_to_swrl/3
	]).

:- use_module(drs_to_owlswrl_core, [
		is_toplevel/3,
		get_namedclass/2,
		get_namedobjectproperty/2,
		get_nameddataproperty/2,
		condlist_and/4,
		is_object_with_generalized_quantifier/1,
		dataitem_datavalue_datatypeuri/3
	]).


/** <module> DRS implication to SWRL 'Implies'-rule translator

Translate an Attempto DRS implication into Semantic Web Rule Language (SWRL) 'Implies'-rule.

TODO: support string concat

@author Kaarel Kaljurand
@version 2010-07-05
*/


%% implication_to_swrl(+Condition:term, +RefList:list, -Implies:term) is semidet.
%
% @param Condition is a DRS implication condition
% @param RefList is a list of DRS toplevel referents
% @param Implies is a SWRL 'Implies'-rule
%
implication_to_swrl(
	'=>'(CondList1, CondList2),
	RefList,
	'Implies'('Antecedent'(AtomList1), 'Consequent'(AtomList2))
) :-
	condlist_atomlist(CondList1, RefList, AtomList1),
	condlist_atomlist(CondList2, RefList, AtomList2).


%% condlist_atomlist(+ConditionList:list, +RefList:list, -AtomList:list) is det.
%
% Note that most conditions can produce several SWRL atoms because
% an argument of the condition can be a complex expression which
% will have to be unravelled into multiple SWRL builtins.
%
% @param ConditionList is a list of DRS conditions
% @param RefList is a list of DRS toplevel referents
% @param AtomList is list of SWRL atoms
%
condlist_atomlist([], _, []).

condlist_atomlist([Condition | ConditionList], RefList, AtomList) :-
	condition_atomlist(Condition, RefList, AtomList1),
	condlist_atomlist(ConditionList, RefList, AtomList2),
	append(AtomList1, AtomList2, AtomList).


/*
% BUG: experimental
% BUG: Add a restriction that all variables used in predicate-conditions
% must be declared in object-conditions.

condlist_atomlist([], _RefList, []).

condlist_atomlist([Condition | CondList], RefList, [Atom | AtomList]) :-
	(
		Condition = object(D, Value, countable,  _, _, _)-_
	->
		get_namedclass(Value, NamedClass),
		select(Condition1, CondList, CondList1),
		condlist_class(D, Condition1, CondList1, RefList, AndClass, CondList2),
		Atom = 'description'(and(NamedClass, AndClass), 'I-variable'(D))
	;
		condition_atom(Condition, RefList, Atom),
		CondList2 = CondList
	),
	condlist_atomlist(CondList2, RefList, AtomList).
*/



%% condition_atomlist(+Condition:term, -Atoms:list) is det.
%
% @param Condition is a DRS condition
% @param RefList is a list of DRS toplevel referents
% @param Atoms is a list of SWRL atoms
%

% We allow no generalized quantifiers
condition_atomlist(Condition, _, _) :-
	is_object_with_generalized_quantifier(Condition),
	!,
	fail.

condition_atomlist(formula(Expr1, Comp, Expr2)-_, RefList, Builtins) :-
	expr_to_builtins(Expr1, Var1, B1),
	expr_to_builtins(Expr2, Var2, B1, B2),
	add_relation(Comp, Var1, Var2, RefList, B2, Builtins).

condition_atomlist(object(X, Value, _,  _, _, _)-_, _, ['description'(NamedClass, 'I-variable'(X))]) :-
	get_namedclass(Value, NamedClass).

condition_atomlist('-'([predicate(_, be, X1, X2)-_]), RefList, ['differentFrom'(OutX1, OutX2)]) :-
	get_argument(X1, RefList, object, OutX1, _),
	get_argument(X2, RefList, object, OutX2, _).

condition_atomlist(predicate(_, be, X1, X2)-_, RefList, ['sameAs'(OutX1, OutX2)]) :-
	get_argument(X1, RefList, object, OutX1, _),
	get_argument(X2, RefList, object, OutX2, _).

condition_atomlist(predicate(_, Value, X1, X2)-_, RefList, ['description'(DataProperty, OutX1, OutX2) | L]) :-
	Value \= be,
	get_nameddataproperty(Value, DataProperty),
	get_argument(X1, RefList, Type1, OutX1, L1),
	get_argument(X2, RefList, Type2, OutX2, L2),
	(
		Type1 = data
	;
		Type2 = data
	),
	append(L1, L2, L).

condition_atomlist(predicate(_, Value, X1, X2)-_, RefList, ['description'(ObjectProperty, OutX1, OutX2)]) :-
	Value \= be,
	get_namedobjectproperty(Value, ObjectProperty),
	get_argument(X1, RefList, object, OutX1, _),
	get_argument(X2, RefList, object, OutX2, _).

% Sublist
% E.g. If a man owns a car then the man likes the car and likes exactly 3 cats.
% E.g. If John owns a car then Mary likes a car and likes at most 3 cats.
condition_atomlist([C | Cs], RefList, ['description'(SubListClass, OutX)]) :-
	condlist_and(X, [C | Cs], RefList, SubListClass),
	get_argument(X, RefList, object, OutX, _).

% Note that X can point to an Individual (i.e. not only a variable),
% see e.g. "If John does not like Mary then a man does not like a woman."
condition_atomlist('-'(Not), RefList, ['description'('ObjectComplementOf'(NotClass), OutX)]) :-
	condlist_and(X, Not, RefList, NotClass),
	get_argument(X, RefList, object, OutX, _).

% Note that X can point to an Individual.
condition_atomlist('v'(Or1, Or2), RefList, ['description'('ObjectUnionOf'([Or1Class, Or2Class]), OutX)]) :-
	condlist_and(X, Or1, RefList, Or1Class),
	condlist_and(X, Or2, RefList, Or2Class),
	get_argument(X, RefList, object, OutX, _).


%% get_argument(+X:term, +RefList:list, -Type:atom, -Argument:term, -Atoms:list) is det.
%
% @param X is a discourse referent, expr-term, int/1, real/1, or string/1
% @param RefList is a list of DRS toplevel referents
% @param Type is the type of the argument, i.e. one of {data, object, _}
% @param Argument is a SWRL variable, OWL individual, or OWL data value
% @param Atoms is a list of SWRL atoms (empty if X produces no builtins)
%
get_argument(expr(X, Y, Z), _RefList, data, Var, Builtins) :-
	!,
	expr_to_builtins(expr(X, Y, Z), Var, Builtins).

% int/1, real/1, string/1
% BUG: use a predicate that already returns the ^^/2 term
% to make sure that this term is used everywhere in DRS->OWL/SWRL
% where numbers are represented.
get_argument(DataItem, _RefList, data, '^^'(DataValue, DataTypeUri), []) :-
	dataitem_datavalue_datatypeuri(DataItem, DataValue, DataTypeUri),
	!.

get_argument(X, RefList, object, Individual, []) :-
	is_toplevel(X, RefList, 'ObjectOneOf'([Individual])),
	!.

get_argument('$VAR'(X), _RefList, object, 'I-variable'(X), []).

get_argument(X, _RefList, object, X, []).


%% add_relation(+Relation:atom, +Var1, +Var2, +BIn:list, -BOut:list) is det.
%
% @param Relation is one of {'>', '<', '='}
% ...
%
% @bug try to output less code, e.g. instead of 'equal' just unify: add_relation('=', Var, Var, B, B).
%
add_relation(Relation, Var1, Var2, RefList, B, [builtIn(Builtin, OutVar1, OutVar2) | B]) :-
	relation_builtin(Relation, Builtin),
	get_argument(Var1, RefList, _, OutVar1, _),
	get_argument(Var2, RefList, _, OutVar2, _).


% expr_to_builtins(+Expr:term, -EList:list) is det.
% expr_to_builtins(+Expr:term, -Var:atom, -EList:list) is det.
%
% Flattens the expression (a tree) into a list simple expressions
% that are connected by expression handles. E.g.
%
%==
% ?- expr_to_builtins(expr(*, expr(+, 1, 2), 3), L).
% L = [builtin(multiply, G295, G297, 3), builtin(add, G297, 1, 2)].
%==
%
% @param Expr is a term in the form expr(Op, Expr1, Expr2)
% @param Var is a variable to be used as handle for the topmost expression
% @param Builtins is a list of SWRL built-in predicates
%
expr_to_builtins(Expr, List) :-
	expr_to_builtins(Expr, _, List).

expr_to_builtins(Expr, Var, List) :-
	expr_to_builtins(Expr, Var, [], List).

expr_to_builtins(expr(Op, E1, E2), Var, LIn, [builtIn(Builtin, Var, Var1, Var2) | LOut]) :-
	!,
	op_builtin(Op, Builtin),
	get_swrl_variable(Var),
	expr_to_builtins(E1, Var1, LIn, LTmp),
	expr_to_builtins(E2, Var2, LTmp, LOut).

% @bug: why is RefList here empty?
expr_to_builtins(Var, OutVar, List, List) :-
	get_argument(Var, [], _, OutVar, _).


%% get_swrl_variable(-SwrlVariable:term) is det.
%
% Generate a new SWRL variable (prefixed by 'g' to avoid a possible nameclash)
%
% @bug It's not nice to use gensym/2
%
get_swrl_variable('I-variable'(Var)) :-
	gensym(g, Var).


%% relation_builtin(?Relation:atom, ?Builtin:atom) is det.
%
% @param Relation is one of {'>', '<', '='}
% @param Builtin is SWRL built-in
%
relation_builtin('=', equal).
relation_builtin('<', lessThan).
relation_builtin('>', greaterThan).


%% op_builtin(?Op:atom, ?Builtin:atom) is det.
%
% @param Op is DRS operator
% @param Builtin is SWRL built-in
%
op_builtin('+', add).
op_builtin('-', subtract).
op_builtin('*', multiply).
op_builtin('/', divide).
